'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const provider_factory_1 = require("./provider-factory");
const completion_providers_1 = require("./completion-providers");
const diagnosis_1 = require("./diagnosis");
const VCL = require("vscode-css-languageservice");
const vscode_languageserver_types_1 = require("vscode-languageserver-types");
const uri_utils_1 = require("./utils/uri-utils");
const provider_1 = require("./provider");
class StylableLanguageService {
    constructor(connection, services, fs, docsDispatcher, notifications) {
        const provider = provider_factory_1.createProvider(services.styl, services.tsLanguageService);
        const processor = provider.styl.fileProcessor;
        const cssService = VCL.getCSSLanguageService();
        connection.onInitialize((params) => {
            return {
                capabilities: {
                    textDocumentSync: 1,
                    completionProvider: {
                        triggerCharacters: ['.', '-', ':', '"', ',']
                    },
                    definitionProvider: true,
                    hoverProvider: true,
                    referencesProvider: true,
                    colorProvider: true,
                    signatureHelpProvider: {
                        triggerCharacters: [
                            '(',
                            ','
                        ]
                    },
                }
            };
        });
        connection.onCompletion((params) => {
            if (!params.textDocument.uri.endsWith('.st.css') && !params.textDocument.uri.startsWith('untitled:')) {
                return Promise.resolve([]);
            }
            const cssCompsRaw = cssService.doComplete(fs.get(params.textDocument.uri), params.position, cssService.parseStylesheet(fs.get(params.textDocument.uri)));
            const cssComps = cssCompsRaw ? cssCompsRaw.items : [];
            const doc = fs.get(params.textDocument.uri).getText();
            const pos = params.position;
            return provider.provideCompletionItemsFromSrc(doc, { line: pos.line, character: pos.character }, params.textDocument.uri, fs)
                .then((res) => {
                return res.map((com) => {
                    let lspCompletion = vscode_languageserver_types_1.CompletionItem.create(com.label);
                    let ted = vscode_languageserver_types_1.TextEdit.replace(com.range ? com.range : new completion_providers_1.ProviderRange(new completion_providers_1.ProviderPosition(pos.line, Math.max(pos.character - 1, 0)), pos), typeof com.insertText === 'string' ? com.insertText : com.insertText.source);
                    lspCompletion.insertTextFormat = 2;
                    lspCompletion.detail = com.detail;
                    lspCompletion.textEdit = ted;
                    lspCompletion.sortText = com.sortText;
                    lspCompletion.filterText = typeof com.insertText === 'string' ? com.insertText : com.insertText.source;
                    if (com.additionalCompletions) {
                        lspCompletion.command = vscode_languageserver_types_1.Command.create("additional", "editor.action.triggerSuggest");
                    }
                    else if (com.triggerSignature) {
                        lspCompletion.command = vscode_languageserver_types_1.Command.create("additional", "editor.action.triggerParameterHints");
                    }
                    return lspCompletion;
                }).concat(cssComps);
            });
        });
        function diagnose(document) {
            let cssDiags = document.uri.endsWith('.css')
                ? cssService.doValidation(document, cssService.parseStylesheet(document))
                    .filter(diag => {
                    if (diag.code === 'emptyRules') {
                        return false;
                    }
                    if (diag.code === 'css-unknownatrule' && readDocRange(document, diag.range) === '@custom-selector') {
                        return false;
                    }
                    if (diag.code === 'css-lcurlyexpected' && readDocRange(document, vscode_languageserver_types_1.Range.create(vscode_languageserver_types_1.Position.create(diag.range.start.line, 0), diag.range.end)).startsWith('@custom-selector')) {
                        return false;
                    }
                    if (diag.code === 'unknownProperties') {
                        let prop = diag.message.match(/'(.*)'/)[1];
                        let src = fs.loadTextFileSync(document.uri);
                        let meta = provider_1.createMeta(src, document.uri).meta;
                        if (meta && Object.keys(meta.mappedSymbols).some(ms => ms === prop)) {
                            return false;
                        }
                    }
                    return true;
                })
                    .map(diag => {
                    diag.source = 'css';
                    return diag;
                })
                : [];
            let diagnostics = diagnosis_1.createDiagnosis(document, fs, processor, services.requireModule).map(diag => { diag.source = 'stylable'; return diag; });
            connection.sendDiagnostics({ uri: document.uri, diagnostics: diagnostics.concat(cssDiags) });
        }
        docsDispatcher.onDidOpen(function (params) {
            diagnose(params.document);
        });
        docsDispatcher.onDidChangeContent(function (params) {
            diagnose(params.document);
        });
        connection.onDefinition((params) => {
            const doc = fs.loadTextFileSync(params.textDocument.uri);
            const pos = params.position;
            return provider.getDefinitionLocation(doc, { line: pos.line, character: pos.character }, uri_utils_1.fromVscodePath(params.textDocument.uri), fs)
                .then((res) => {
                return res.map(loc => vscode_languageserver_types_1.Location.create(uri_utils_1.toVscodePath(loc.uri), loc.range));
            });
        });
        connection.onHover((params) => {
            let hover = cssService.doHover(fs.get(params.textDocument.uri), params.position, cssService.parseStylesheet(fs.get(params.textDocument.uri)));
            return Promise.resolve(hover); //Because Hover may be null, but not of null type - vscode code isn't strict that way.
        });
        connection.onReferences((params) => {
            let refs = cssService.findReferences(fs.get(params.textDocument.uri), params.position, cssService.parseStylesheet(fs.get(params.textDocument.uri)));
            return Promise.resolve(refs);
        });
        connection.onRequest(notifications.colorRequest.type, params => {
            const document = fs.get(params.textDocument.uri);
            const src = document.getText();
            const stylesheet = cssService.parseStylesheet(document);
            const colors = cssService.findDocumentColors(document, stylesheet);
            return colors;
        });
        connection.onRequest(notifications.colorPresentationRequest.type, params => {
            const document = fs.get(params.textDocument.uri);
            const stylesheet = cssService.parseStylesheet(document);
            const colors = cssService.getColorPresentations(document, stylesheet, params.color, params.range);
            return colors;
        });
        connection.onSignatureHelp((params) => {
            const doc = fs.loadTextFileSync(params.textDocument.uri);
            let sig = provider.getSignatureHelp(doc, params.position, params.textDocument.uri, fs, vscode_languageserver_types_1.ParameterInformation);
            return Promise.resolve(sig);
        });
        function readDocRange(doc, rng) {
            let lines = doc.getText().split('\n');
            return lines[rng.start.line].slice(rng.start.character, rng.end.character);
        }
    }
}
exports.StylableLanguageService = StylableLanguageService;
//# sourceMappingURL=service.js.map